#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>

struct char_freq {
	char who;
	double when;
};

struct model {
	struct char_freq items[256];
};

#include "models.h"

#define lenof(obj) (sizeof(obj) / sizeof(*(obj)))

void halt_and_catch() {
	fprintf(stderr, "critical error occurred. please leave the building immediately\n");
	exit(1);
}

char *readfile(FILE *file, size_t *size) {
	int cap = 16;
	char *buf = malloc(cap);
	if (!buf) return NULL;
	int count = 0, c;
	while ((c=fgetc(file))!=EOF) {
		count++;
		if (count > cap) {
			char *buf_ = buf;
			if (!(buf = realloc(buf, (cap *= 2)))) {
				free(buf_);
				return NULL;
			}
		}
		buf[count-1] = c;
	}
	char *buf_ = buf;
	if (!(buf = realloc(buf, count))) {
		free(buf_);
		return NULL;
	}
	*size = count;
	return buf;
}

int sort_model_cb(const void *a_, const void *b_) {
	const struct char_freq *a = a_;
	const struct char_freq *b = b_;
	double diff = b->when - a->when;
	return diff < 0 ? -1 : (diff > 0 ? 1 : 0);
}

void sort_model(struct model *model) {
	qsort(model->items, 256, sizeof(*model->items), sort_model_cb);
}

double cmp_model(struct model *a, struct model *b) {
	double accum = 0;
	for (int i = 0; i < 256; i++) {
		double diff = (a->items[i].when - b->items[i].when);
		accum += diff * diff;
	}
	return accum;
}

void merge_model(const struct model *a, double a_weight, const struct model *b, struct model *out) {
	if (a_weight == 0) {
		*out = *b;
		return;
	}
	struct model b_raw;
	for (int i = 0; i < 256; i++) {
		b_raw.items[(unsigned char)b->items[i].who] = b->items[i];
	}
	for (int i = 0; i < 256; i++) {
		out->items[i] = (struct char_freq) {
			.who = a->items[i].who, .when =
				(a->items[i].when
				 + b_raw.items[(unsigned char)a->items[i].who].when)
				 / (a_weight+1)
		};
	}
	sort_model(out);
}

int analyze_buffer(const char *str, size_t size) {
	size_t freqs[256] = {0};
	for (int i = 0; i < size; i++) {
		freqs[(unsigned char)str[i]]++;
	}
	struct model model;
	for (int i = 0; i < 256; i++) {
		model.items[i] = (struct char_freq) {
			.who = i, .when = ((double)freqs[i])/size,
		};
	}
	sort_model(&model);
	double bestest_score = 1.0/0.0;
	struct model bestest_model;
	for (int d = 0 ;; d++) {
		double best_model_score = 1.0/0.0;
		struct model best_model;
		struct model new_model;
		for (int i = 0; i < lenof(models); i++) {
			merge_model(&bestest_model, d, &models[i], &new_model);
			double score = cmp_model(&model, &new_model);
			if (score < best_model_score) {
				best_model_score = score;
				best_model = new_model;
			}
		}
		if (best_model_score >= bestest_score) {
			break;
		}
		bestest_score = best_model_score;
		bestest_model = best_model;
	}
	char map[256];
	for (int i = 0; i < 256; i++) {
		map[(unsigned char)model.items[i].who] = bestest_model.items[i].who;
	}
	for (int i = 0; i < size; i++) {
		fputc(map[(unsigned char)str[i]], stdout);
	}
	return 0;
}

int analyze_file(FILE *file) {
	size_t size;
	char *data = readfile(file, &size);
	if (!(data && !analyze_buffer(data, size))) {
		return 1;
	}
	return 0;
}

int main(int argc, char **argv) {
	FILE *file = stdin;
	if(argc > 1) {
		file = fopen(argv[1], "rb");
		if (!file) {
			halt_and_catch();
		}
	}
	if (analyze_file(file)) {
		halt_and_catch();
	}
	return 0;
}
